<Query Kind="Program">
  <Namespace>System.Threading.Tasks</Namespace>
  <Namespace>System.Net</Namespace>
</Query>

public static class Extensions
{
	public static Task<T> Catch<T,TError> (this Task<T> task, Func<TError,T> onError) 
		where TError : Exception
	{
		var tcs = new TaskCompletionSource<T> ();
		
		task.ContinueWith (ant =>
		{
			if (task.IsFaulted && task.Exception.InnerException is TError)
				tcs.SetResult (onError ((TError) task.Exception.InnerException));
			else if (ant.IsCanceled)
				tcs.SetCanceled ();
			else if (task.IsFaulted)
				tcs.SetException (ant.Exception.InnerException);
			else
				tcs.SetResult (ant.Result);
		});
		return tcs.Task;
	}	
	
#region Simple error value

	public static Task<T> Catch<T,TError> (this Task<T> task, T onError) where TError : Exception
		=> task.Catch<T,TError> (ex => onError);
		
#endregion
#region Nongeneric Task	

	public static Task<T> Catch<T,TError> (this Task task, T onError) where TError : Exception
		=> task.ToTaskOfT<T>().Catch<T,TError> (onError);

#endregion
#region ToTaskOfT

	public async static Task<T> ToTaskOfT<T> (this Task t)
	{
		await t.ConfigureAwait (false);
		return default(T);
	}

#endregion
#region ToTaskOfT - optimized

	public static Task<T> ToTaskOfT_<T> (this Task t)
	{
		if (t is Task<T>) return (Task<T>)t;
		var tcs = new TaskCompletionSource<T> ();
		t.ContinueWith (ant =>
		{
			if (ant.IsCanceled) tcs.SetCanceled ();
			else if (ant.IsFaulted) tcs.SetException (ant.Exception.InnerException);
			else tcs.SetResult (default (T));
		});
		return tcs.Task;
	}
	
#endregion
#region 'Swallow' exceptions

	public static Task<T> Catch<T> (this Task<T> task, T valueIfError = default(T))
		=> task.Catch<T,Exception> (valueIfError);
	
	public static Task Catch (this Task task)
		=> task.Catch<object,Exception> (null);

#endregion
#region Timeouts and cancellation

	public static Task<T> OnTimeout<T> (this Task<T> t, T valueIfError)
		=> t.Catch<T,TimeoutException> (valueIfError);

	public static Task<T> OnTimeout<T> (this Task t, T valueIfError)
		=> OnTimeout (ToTaskOfT<T> (t), valueIfError);

	public static Task<T> OnCancellation<T> (this Task<T> t, T valueIfError)
		=> t.Catch<T,OperationCanceledException> (valueIfError);

	public static Task<T> OnCancellation<T> (this Task t, T valueIfError)
		=> OnCancellation (ToTaskOfT<T> (t), valueIfError);

#endregion
}

#region Demo

async Task<int> SomeAsyncMethod()
{
	#region implementation
	await Task.Delay(1000);
	return 123;
	#endregion
}

async void NonFluent()
{
	int? result;
	try
	{
		result = await SomeAsyncMethod();
	}
	catch (TimeoutException)
	{
		result = null;
	}
	catch (OperationCanceledException)
	{
		result = null;
	}
}

async void Fluent()
{
	int? result = await SomeAsyncMethod()
		.OnTimeout<int?>(null)
		.OnCancellation (null);
}

#endregion